#include "petsc.h"
#include "petscdmplex.h"
#include "petscdm.h"
#include "petscdmlabel.h"
#include "petscviewerhdf5.h"
#include "petscds.h"
#include "petscvec.h"
#include <time.h>
#include <stdlib.h>

int main(int argc, char **argv)
{

/**** Generate Periodic DM ******/
const PetscBool  cellSimplex = PETSC_FALSE;
const PetscBool  interpBefore = PETSC_TRUE;
PetscInt 		 dim = 3;
PetscInt 		 iSize[3];
PetscInt         i, numPoints;
PetscInt         overlap = 1, numFields = 1;
PetscScalar 	 Xmin[3], Xmax[3];
PetscScalar     *xVecArray;
DMBoundaryType 	 MeshPeriodicity[3]; /* Periodicity of mesh in each direction */
DM 				 dm, pdm = NULL;
PetscPartitioner part;
PetscErrorCode	 ierr;

srand(time(NULL));   /* Random values initialization */

ierr = PetscInitialize(&argc, &argv, NULL, NULL);
CHKERRQ(ierr);

// Number of cells in each direction
iSize[0] = 10;
iSize[1] = 10;
iSize[2] = 10;

// Dimensions
Xmin[0] = 0.0;
Xmin[1] = 0.0;
Xmin[2] = 0.0;
Xmax[0] = 1.0;
Xmax[1] = 1.0;
Xmax[2] = 1.0;

// Periodicity
MeshPeriodicity[0] = DM_BOUNDARY_PERIODIC;
MeshPeriodicity[1] = DM_BOUNDARY_PERIODIC;
MeshPeriodicity[2] = DM_BOUNDARY_PERIODIC;
//MeshPeriodicity[0] = DM_BOUNDARY_NONE;
//MeshPeriodicity[1] = DM_BOUNDARY_NONE;
//MeshPeriodicity[2] = DM_BOUNDARY_NONE;

// Create Mesh
ierr = DMPlexCreateBoxMesh(PETSC_COMM_WORLD,dim,cellSimplex,iSize,Xmin,Xmax,MeshPeriodicity,interpBefore,&dm);
CHKERRQ(ierr);

// Create Partition
ierr = DMPlexGetPartitioner(dm, &part);
CHKERRQ(ierr);
if (part)
{
  ierr = PetscPartitionerSetType(part, PETSCPARTITIONERPARMETIS);
  CHKERRQ(ierr);
  ierr = PetscPartitionerSetFromOptions(part);
  CHKERRQ(ierr);
}

// Distribute
ierr = DMPlexDistribute(dm, overlap, NULL, &pdm);
CHKERRQ(ierr);
if (pdm)
{
	  ierr = DMDestroy(&dm);
	  CHKERRQ(ierr);
	  dm = pdm;
	  ierr = PetscObjectSetName((PetscObject) dm, "Distributed Mesh");
	  CHKERRQ(ierr);
 }

ierr = DMLocalizeCoordinates(dm);
CHKERRQ(ierr);
ierr = DMSetFromOptions(dm);
CHKERRQ(ierr);

// Create Section on DM
PetscSection section;
PetscInt *numComp;
PetscInt *numDof;

ierr = PetscCalloc2(numFields, &numComp, numFields * (dim + 1), &numDof);
CHKERRQ(ierr);

for (i = 0; i < numFields; i++)
{
  numComp[i] = 1;
  numDof[i * (dim + 1) + dim] = 1;
}

ierr = DMSetNumFields(dm, numFields);
CHKERRQ(ierr);
ierr = DMPlexCreateSection(dm, NULL, numComp, numDof, 0, NULL, NULL, NULL, NULL, &section);
CHKERRQ(ierr);

// Push section onto the DM
ierr = DMSetLocalSection(dm, section);
CHKERRQ(ierr);
ierr = PetscFree2(numComp, numDof);
CHKERRQ(ierr);
ierr = PetscSectionDestroy(&section);
CHKERRQ(ierr);

/**** Create Vector Field => xGlobalVector *********/
Vec xGlobalVector;

ierr = DMCreateGlobalVector(dm, &xGlobalVector);
CHKERRQ(ierr);

// Initialize the Vec to 0
ierr = VecSet(xGlobalVector,0.0);
CHKERRQ(ierr);

// Fill the Vector with random values
ierr = VecGetArray(xGlobalVector, &xVecArray);
CHKERRQ(ierr);
ierr = VecGetLocalSize(xGlobalVector, &numPoints);
CHKERRQ(ierr);

for (i = 0; i < numPoints; i++) /* Loop over all internal cells */
{
	 xVecArray[i] = rand() % 20;
}

ierr = VecRestoreArray(xGlobalVector, &xVecArray);
CHKERRQ(ierr);

/**** Write DM + associated Vec ****/
PetscViewer H5Viewer;
ierr = PetscViewerHDF5Open(PETSC_COMM_WORLD, "periodic_box.h5", FILE_MODE_WRITE, &H5Viewer);
CHKERRQ(ierr);
ierr = PetscObjectSetName((PetscObject)dm, "plexA");
CHKERRQ(ierr);
ierr = PetscViewerPushFormat(H5Viewer, PETSC_VIEWER_HDF5_PETSC);
CHKERRQ(ierr);
ierr = DMPlexTopologyView(dm, H5Viewer);
CHKERRQ(ierr);
ierr = DMPlexLabelsView(dm, H5Viewer);
CHKERRQ(ierr);
ierr = DMPlexCoordinatesView(dm, H5Viewer);
CHKERRQ(ierr);
ierr = PetscViewerPopFormat(H5Viewer);
CHKERRQ(ierr);

DM sdm;
PetscSection s;

ierr = DMClone(dm, &sdm);
CHKERRQ(ierr);
ierr = PetscObjectSetName((PetscObject)sdm, "dmA");
CHKERRQ(ierr);
ierr = DMGetGlobalSection(dm, &s);
CHKERRQ(ierr);
ierr = DMSetGlobalSection(sdm, s);
CHKERRQ(ierr);
ierr = DMPlexSectionView(dm, H5Viewer, sdm);
CHKERRQ(ierr);

Vec  vec;
PetscScalar *array;

ierr = DMGetGlobalVector(sdm, &vec);
CHKERRQ(ierr);

/*** Fill the vector xGlobalVector ***/
ierr = VecGetArray(vec, &array);
CHKERRQ(ierr);
ierr = VecGetLocalSize(xGlobalVector, &numPoints);
CHKERRQ(ierr);
ierr = VecGetArray(xGlobalVector, &xVecArray);
CHKERRQ(ierr);

for (i = 0; i < numPoints; i++) /* Loop over all internal cells */
{
      array[i] = xVecArray[i];
}

ierr = VecRestoreArray(vec, &array);
CHKERRQ(ierr);
ierr = VecRestoreArray(xGlobalVector, &xVecArray);
CHKERRQ(ierr);

ierr = PetscObjectSetName((PetscObject)vec, "vecA");
CHKERRQ(ierr);
ierr = DMPlexGlobalVectorView(dm, H5Viewer, sdm, vec);
CHKERRQ(ierr);
ierr = PetscViewerDestroy(&H5Viewer);
CHKERRQ(ierr);
/*** End of writing ****/

/*** Wait for all CPUs to end ***/
ierr = MPI_Barrier(PETSC_COMM_WORLD);
CHKERRQ(ierr);

/*** Load ***/
ierr = DMDestroy(&dm);
CHKERRQ(ierr);
ierr = DMDestroy(&sdm);
CHKERRQ(ierr);

Vec Restart_xGlobalVector;
PetscSF sfO, sf, sfDist, globalDataSF, localDataSF;
DM distributedDM;

ierr = PetscViewerHDF5Open(PETSC_COMM_WORLD, "periodic_box.h5", FILE_MODE_READ, &H5Viewer);
CHKERRQ(ierr);
ierr = DMCreate(PETSC_COMM_WORLD, &dm);
CHKERRQ(ierr);
ierr = DMSetType(dm, DMPLEX);
CHKERRQ(ierr);
ierr = PetscObjectSetName((PetscObject)dm, "plexA");
CHKERRQ(ierr);
ierr = PetscViewerPushFormat(H5Viewer, PETSC_VIEWER_HDF5_PETSC);
CHKERRQ(ierr);
ierr = DMPlexTopologyLoad(dm, H5Viewer, &sfO);
CHKERRQ(ierr);
ierr = DMPlexLabelsLoad(dm, H5Viewer);
CHKERRQ(ierr);
ierr = DMPlexCoordinatesLoad(dm, H5Viewer, sfO);
CHKERRQ(ierr);
ierr = PetscViewerPopFormat(H5Viewer);
CHKERRQ(ierr);

ierr = DMPlexDistribute(dm, overlap, &sfDist, &distributedDM);
CHKERRQ(ierr);
if (distributedDM) {
      ierr = DMDestroy(&dm);
      CHKERRQ(ierr);
      dm = distributedDM;
      ierr = PetscObjectSetName((PetscObject)dm, "plexA");
      CHKERRQ(ierr);
}

ierr = PetscSFCompose(sfO, sfDist, &sf);
CHKERRQ(ierr);
ierr = PetscSFDestroy(&sfO);
CHKERRQ(ierr);
ierr = PetscSFDestroy(&sfDist);
CHKERRQ(ierr);

ierr = DMClone(dm, &sdm);
CHKERRQ(ierr);
ierr = PetscObjectSetName((PetscObject)sdm, "dmA");
CHKERRQ(ierr);
ierr = DMPlexSectionLoad(dm, H5Viewer, sdm, sf, &globalDataSF, &localDataSF);
CHKERRQ(ierr);

/*** Load the Vector to Restart_xGlobalVector ***/
ierr = DMGetGlobalVector(sdm, &Restart_xGlobalVector);
CHKERRQ(ierr);
ierr = VecSet(Restart_xGlobalVector,0.0);
CHKERRQ(ierr);

ierr = PetscObjectSetName((PetscObject)Restart_xGlobalVector, "vecA");
CHKERRQ(ierr);
ierr = DMPlexGlobalVectorLoad(dm, H5Viewer, sdm,globalDataSF,Restart_xGlobalVector);
CHKERRQ(ierr);

ierr = PetscViewerDestroy(&H5Viewer);
CHKERRQ(ierr);

PetscEnd();
}

// *************** Write ***************
//PetscInt rank;
//ierr = MPI_Comm_rank(MPI_COMM_WORLD, &rank);
//CHKERRQ(ierr);
//
//if(rank == 0)
//{
//	for (i = 0; i < numPoints; i++)
//	{
//		printf(" Write %e %i, %i \n", xVecArray[i],i, rank);
//	}
//}
//
//ierr = MPI_Barrier(PETSC_COMM_WORLD);
//CHKERRQ(ierr);
//
//if(rank == 1)
//{
//	for (i = 0; i < numPoints; i++)
//	{
//		printf(" Write %e %i, %i \n", xVecArray[i],i, rank);
//	}
//}

// ***************** LOAD *******************
//ierr = VecGetArray(Restart_xGlobalVector, &xVecArray);
//CHKERRQ(ierr);
//ierr = VecGetLocalSize(Restart_xGlobalVector, &numPoints);
//CHKERRQ(ierr);
//
//if(rank == 0)
//{
//	for (i = 0; i < numPoints; i++)
//	{
//		printf(" LOAD %e %i, %i \n", xVecArray[i],i, rank);
//	}
//}
//
//ierr = MPI_Barrier(PETSC_COMM_WORLD);
//CHKERRQ(ierr);
//
//if(rank == 1)
//{
//	for (i = 0; i < numPoints; i++)
//	{
//		printf(" LOAD %e %i, %i \n", xVecArray[i],i, rank);
//	}
//}
//ierr = VecRestoreArray(Restart_xGlobalVector, &xVecArray);
//CHKERRQ(ierr);
